def binary_mask(raster_path , shape_path, output_path, file_name ):
    
    """Function that generates a binary mask from a vector file (shp)
    
    Parameters
    ----------
    
    raster_path = path to the .tif;

    shape_path = path to the shapefile or GeoJson.

    output_path = Path to save the binary mask.

    file_name = Name of the file.
    
    Returns
    -------
    Raster Binary Mask in tif format
    
    """
    import rasterio
    from rasterio.features import rasterize
    import geopandas as gpd
    import numpy as np
    import os
    
    if not os.path.exists( output_path):
        os.makedirs( output_path)

    # Load the shapefile into a GeoDataFrame
    shapefile = gpd.read_file(shape_path)

    # Load the raster that you want to use as the base for the mask
    with rasterio.open(raster_path) as src:
        transform = src.transform
        mask = src.read(1)
        mask = mask.astype('float32')
        mask[mask == src.nodata] = 0

    # Create a binary mask using the shapefile
    mask = mask.copy()
    mask[:] = 1
    for shape in shapefile.geometry:
        mask = rasterize(
            [(shape, 0)],
            out_shape=mask.shape,
            transform=transform,
            fill=1,
            all_touched=True,
            default_value=0,
            dtype=np.uint8
        )

    # Save the binary mask as a TIFF file
    with rasterio.open(output_path+"/" + file_name+".tif", "w", driver="GTiff", height=mask.shape[0], width=mask.shape[1],
                    count=1, dtype=str(mask.dtype), crs=src.crs, transform=transform, nodata=0) as dst:
        dst.write(mask.astype(mask.dtype), 1)


def DynamicChangeDetection(Path_working_Directory=r"" , Path_UDM2_folder=r"",
 Path_to_DEMFile=r"", Coh_Thresh=0.75 , vel_thresh=0.063 , udm_mask_option=False , cmap='jet', Median_Filter=False, Set_fig_MinMax=False, show_figure=False, plot_option="origional", xres=10, yres=10):
    

    """
    This program calculates optical flow velocity from triplets of daily optical satellite images.
    Final Timeseris products will be a shapefile format using Time_Series function after stackprep step.
    
    Parameters
    ----------

    Path_working_Directory : str

    Path_UDM2_folder : str

    Path_to_DEMFile : str

    Coh_Thresh : float

    vel_thresh : float

    udm_mask_option : bool

    cmap : str

    Median_Filter : bool

    Set_fig_MinMax : bool

    show_figure : bool

    plot_option : str
        "origional",  "resampled"
    
    xres : int

    yres : int

    Returns
    -------
    Rasters
         velocity in X direction(EW)
         Velocity in Y direction(NS)

    Figures  
        Initial Timesereis Figures (those figures are only intermediate products needs calibration)
    

    """
    
    def binary_mask(raster_path , shape_path, output_path, file_name ):
    
        import rasterio
        from rasterio.features import rasterize
        import geopandas as gpd
        import numpy as np
        import os
        
        if not os.path.exists( output_path):
            os.makedirs( output_path)

        # Load the shapefile into a GeoDataFrame
        shapefile = gpd.read_file(shape_path)

        # Load the raster that you want to use as the base for the mask
        with rasterio.open(raster_path) as src:
            transform = src.transform
            mask = src.read(1)
            mask = mask.astype('float32')
            mask[mask == src.nodata] = 0

        # Create a binary mask using the shapefile
        mask = mask.copy()
        mask[:] = 1
        for shape in shapefile.geometry:
            mask = rasterize(
                [(shape, 0)],
                out_shape=mask.shape,
                transform=transform,
                fill=1,
                all_touched=True,
                default_value=0,
                dtype=np.uint8
            )

        # Save the binary mask as a TIFF file
        binary_mask_file=output_path+"/" + file_name+".tif"
        with rasterio.open(binary_mask_file, "w", driver="GTiff", height=mask.shape[0], width=mask.shape[1],
                        count=1, dtype=str(mask.dtype), crs=src.crs, transform=transform, nodata=0) as dst:
            dst.write(mask.astype(mask.dtype), 1)

        return binary_mask_file
    import time
    import os
    import rasterio
    import pandas as pd
    from skimage.metrics import structural_similarity as ssim
    import matplotlib.pyplot as plt
    from skimage.registration import optical_flow_tvl1, optical_flow_ilk
    import numpy as np
    from scipy.ndimage import shift
    from pystackreg import StackReg
    from skimage.registration import phase_cross_correlation
    from skimage import exposure
    import cv2
    import os
    from os import listdir
    from os.path import isfile, join
    import numpy as np
    import seaborn_image as sb
    import skimage.transform as st 
    from scipy import interpolate
    from osgeo import gdal
    import numpy as np
    import matplotlib.pyplot as plt
    import cmocean
    import matplotlib.colors as mcolors
    import matplotlib.colors as colors
    from mpl_toolkits.axes_grid1.inset_locator import inset_axes
    import sys
    import warnings
    
    if not sys.warnoptions:
        warnings.simplefilter("ignore")
    
    
    if not os.path.exists( 'georeferenced_folder/flow_xn'):
        os.makedirs( 'georeferenced_folder/flow_xn')
    if not os.path.exists( 'georeferenced_folder/flow_yn'):
        os.makedirs( 'georeferenced_folder/flow_yn')
   
    if not os.path.exists("georeferenced_folder/VEL_Triplets"):
        os.makedirs("georeferenced_folder/VEL_Triplets") 

    

    
            
    # if not os.path.exists("Results_OpticalFlow\gif_dir\georeferenced_folder\direction"):
    #     os.makedirs("Results_OpticalFlow\gif_dir\georeferenced_folder\direction")
           
    output_dirflowxn=r"georeferenced_folder/flow_xn"
    output_dirflowyn=r"georeferenced_folder/flow_yn"
    output_VEL_Triplets=r"georeferenced_folder/VEL_Triplets"
    
    
    # output_dir = r"Results_OpticalFlow"
    # gif_dir= r'Results_OpticalFlow/gif_dir'
    # img_list=glob.glob(Path_working_Directory+"/" + "*.tif")
    # udm2_mask_list=glob.glob(Path_UDM2_folder + "/"+ "*.tif") f for f in sorted(os.listdir(RGBFolder_jpg)) if isfile(join(RGBFolder_jpg, f))
    
    img_list = [f for f in sorted(os.listdir(Path_working_Directory)) if isfile(join(Path_working_Directory, f))]
    
    if udm_mask_option==True:
    	udm2_mask_list=[f for f in sorted(os.listdir(Path_UDM2_folder)) if isfile(join(Path_UDM2_folder, f))]
    
     ####################################################
        #Remove outliers


    def is_outlier(points, thresh=1):
        """
        Returns a boolean array with True if points are outliers and False 
        otherwise.

        Parameters:
        -----------
            points : An numobservations by numdimensions array of observations
            thresh : The modified z-score to use as a threshold. Observations with
                a modified z-score (based on the median absolute deviation) greater
                than this value will be classified as outliers.

        Returns:
        --------
            mask : A numobservations-length boolean array.

        References:
        ----------
            Boris Iglewicz and David Hoaglin (1993), "Volume 16: How to Detect and
            Handle Outliers", The ASQC Basic References in Quality Control:
            Statistical Techniques, Edward F. Mykytka, Ph.D., Editor. 
        """
        if len(points.shape) == 3.5:
            points = points[:,None]
        median = np.median(points, axis=0)
        diff = np.sum((points - median)**2, axis=-1)
        diff = np.sqrt(diff)
        med_abs_deviation = np.median(diff)

        modified_z_score = 0.6745 * diff / med_abs_deviation

        return np.where(modified_z_score > thresh, True, False) #modified_z_score > thresh 

    
    f=1
    for n in range(0, len(img_list)):
        # for item2 in img_list[n+1:]:
        #     for item3 in img_list[n+2:]:
        
        item1=img_list[n]
        item2=img_list[n+1]
        item3=img_list[n+2]
        img_src1=rasterio.open(join(Path_working_Directory,img_list[n]), "r+", masked=True)
        img_src2=rasterio.open(join(Path_working_Directory,img_list[n+1]), "r+", masked=True)
        img_src3=rasterio.open(join(Path_working_Directory,img_list[n+2]), "r+", masked=True)

        img1=img_src1.read(1)
        img2=img_src2.read(1)
        img3=img_src3.read(1)
        
        

        if udm_mask_option==True:
            udm2_src1=rasterio.open(join(Path_UDM2_folder,udm2_mask_list[n]))
            udm2_src2=rasterio.open(join(Path_UDM2_folder,udm2_mask_list[n+1]))
            udm2_src3=rasterio.open(join(Path_UDM2_folder,udm2_mask_list[n+2]))
            udm_img1=udm2_src1.read(1)
            udm_img2=udm2_src2.read(1)
            udm_img3=udm2_src3.read(1)

        

        
        # if udm_mask_option==False:
        #     udm_img1[udm_img1==0]=1
        #     udm_img2[udm_img2==0]=1
        #     udm_img3[udm_img3==0]=1

            img1[udm_img1==0]=0
            img2[udm_img1==0]=0
            img3[udm_img1==0]=0

            img1[udm_img2==0]=0
            img2[udm_img2==0]=0
            img3[udm_img2==0]=0

            img1[udm_img3==0]=0
            img2[udm_img3==0]=0
            img3[udm_img3==0]=0
        # else:
        #     udm_img1=udm_img1
        #     udm_img2=udm_img2
        #     udm_img3=udm_img3


        
       
        bin_mask_img1=binary_mask(raster_path=join(Path_working_Directory,img_list[n]), shape_path='AOI.shp', output_path='bin_mask', file_name=item1)
        bin_mask_img2=binary_mask(raster_path=join(Path_working_Directory,img_list[n+1]), shape_path='AOI.shp', output_path='bin_mask', file_name=item2)
        bin_mask_img3=binary_mask(raster_path=join(Path_working_Directory,img_list[n+2]), shape_path='AOI.shp', output_path='bin_mask', file_name=item3)
        
        img1[bin_mask_img2==0]=0
        img1[bin_mask_img3==0]=0

        img2[bin_mask_img1==0]=0
        img2[bin_mask_img3==0]=0

        img3[bin_mask_img1==0]=0
        img3[bin_mask_img2==0]=0
        
        

        ####Creating Image Title and labels based on YYYYDDMM and HHmm
        # filepath1, img1_name = os.path.split(item1)
        # filepath2, img2_name = os.path.split(item2)
        # filepath3, img3_name = os.path.split(item3)

        img1_name = item1
        img2_name = item2
        img3_name = item3

        #2022-08-12.tif
        Date1_YYYY=img1_name[:-8]
        Date1_MM=img1_name[4:-6]
        Date1_DD=img1_name[6:-4]

        Date2_YYYY=img2_name[:-8]
        Date2_MM=img2_name[4:-6]
        Date2_DD=img2_name[6:-4]

        Date3_YYYY=img3_name[:-8]
        Date3_MM=img3_name[4:-6]
        Date3_DD=img3_name[6:-4]



        #convert dates to number of days in the year for image1
        YMD= Date1_YYYY+Date1_MM+Date1_DD
        date1 = pd.to_datetime(YMD, format='%Y%m%d')
        new_year_day = pd.Timestamp(year=date1.year, month=1, day=1)
        day_of_the_year_date1 = (date1 - new_year_day).days + 1

        #convert dates to number of days in the year for image2
        YMD= Date2_YYYY+Date2_MM+Date2_DD
        date2 = pd.to_datetime(YMD, format='%Y%m%d')
        new_year_day = pd.Timestamp(year=date2.year, month=1, day=1)
        day_of_the_year_date2 = (date2 - new_year_day).days + 1

            #convert dates to number of days in the year for image3
        YMD= Date3_YYYY+Date3_MM+Date3_DD
        date3 = pd.to_datetime(YMD, format='%Y%m%d')
        new_year_day = pd.Timestamp(year=date3.year, month=1, day=1)
        day_of_the_year_date3 = (date3 - new_year_day).days + 1

        Delta_DD= (date3-date1).days
        Delta_DD=int(Delta_DD)
        if Delta_DD < 0:
            Delta_DD=Delta_DD*-1
        else:
            Delta_DD=Delta_DD*1

        print ("day_of_the_year_date1: ", day_of_the_year_date1)
        print("---------------------")
        print ("day_of_the_year_date2: ", day_of_the_year_date2)
        print("---------------------")
        print ("day_of_the_year_date3: ", day_of_the_year_date3)
        print("---------------------")
        print('Delta_DD: '+ str(Delta_DD))

        ##Image resgistration
        # #Affine Body transformation First attempt for image alignement using stackreg affine method
        sr = StackReg(StackReg.AFFINE)
        img2 = sr.register_transform(img1, img2)

        
        img1[bin_mask_img2==0]=0
        img1[bin_mask_img3==0]=0

        img2[bin_mask_img1==0]=0
        img2[bin_mask_img3==0]=0

        img3[bin_mask_img1==0]=0
        img3[bin_mask_img2==0]=0

        img3 = sr.register_transform(img2, img3)

        img1[bin_mask_img2==0]=0
        img1[bin_mask_img3==0]=0

        img2[bin_mask_img1==0]=0
        img2[bin_mask_img3==0]=0

        img3[bin_mask_img1==0]=0
        img3[bin_mask_img2==0]=0

        if udm_mask_option==True:
            
            img1[udm_img1==0]=0
            img2[udm_img1==0]=0
            img3[udm_img1==0]=0

            img1[udm_img2==0]=0
            img2[udm_img2==0]=0
            img3[udm_img2==0]=0

            img1[udm_img3==0]=0
            img2[udm_img3==0]=0
            img3[udm_img3==0]=0
        
        # Registration of the two images: image2 to image1
        shifts12, error, phasediff = phase_cross_correlation(img1, img2, upsample_factor=20, overlap_ratio=3)
        
        img2 = shift(img2, shift=(shifts12[0], shifts12[1]), mode='constant', prefilter=True)

        if udm_mask_option==True:
        
            img1[udm_img1==0]=0
            img2[udm_img1==0]=0
            img3[udm_img1==0]=0

            img1[udm_img2==0]=0
            img2[udm_img2==0]=0
            img3[udm_img2==0]=0

            img1[udm_img3==0]=0
            img2[udm_img3==0]=0
            img3[udm_img3==0]=0

        print ('shift:im1 and im2',shifts12, 'phasediff:', phasediff, 'error:', error )
        
        # Registration of the two images: image3 to image1
        shifts13, error, phasediff = phase_cross_correlation(img2, img3, upsample_factor=20, overlap_ratio=3)
        img3 = shift(img3, shift=(shifts13[0], shifts13[1]), mode='constant', prefilter=True)
        print ('shift:im1 and im3',shifts13, 'phasediff:', phasediff, 'error:', error )

        #Run similarity Threshold Map
        mssim, grad, S12 = ssim(img1, img2, gradient=True, 
        full=True, use_sample_covariance=True, gaussian_weights=True )

        mssim, grad, S13 = ssim(img1, img3, gradient=True, 
        full=True, use_sample_covariance=True, gaussian_weights=True)

        # fig, ax = plt.subplots()
        # img = ax.imshow(S13)
        # fig.colorbar(img, ax=ax) 
        # plt.show()

        #Apply optical Flow 

        

        img1[bin_mask_img2==0]=0
        img1[bin_mask_img3==0]=0

        img2[bin_mask_img1==0]=0
        img2[bin_mask_img3==0]=0

        img3[bin_mask_img1==0]=0
        img3[bin_mask_img2==0]=0

        if udm_mask_option==True:

            img1[udm_img1==0]=0
            img2[udm_img1==0]=0
            img3[udm_img1==0]=0

            img1[udm_img2==0]=0
            img2[udm_img2==0]=0
            img3[udm_img2==0]=0

            img1[udm_img3==0]=0
            img2[udm_img3==0]=0
            img3[udm_img3==0]=0



        # flow1 = optical_flow_tvl1(img1, img2, attachment=15, tightness=0.3, num_warp=5, num_iter=10,
        #                     tol=0.0001, prefilter=True, dtype='float32')
        # flow2 = optical_flow_tvl1(img1, img3,attachment=15, tightness=0.3, num_warp=5, num_iter=10,
        #                     tol=0.0001, prefilter=True, dtype='float32')

    
        # '''

        # 0.4- image pyramid or simple image scale
        # 1 is the number of pyramid layers. 1 means that flow is calculated only from the previous image. 
        # 12 is window size. Flow is computed over the window larger value is more robust to the noise. 
        # 2 mean number of iteration of the algorithm
        # 8 is polynomial degree expansion recommended value is 5 - 7
        # 1.2 standard deviation used to smooth used derivatives recommended values from 1.1 - 1,5


        # '''
        flow1 = cv2.calcOpticalFlowFarneback(img1, img2, flow=None, pyr_scale=0.5, levels=5, winsize=12,iterations= 2, poly_n=7, poly_sigma= 1.5, flags=0)
        flow2= cv2.calcOpticalFlowFarneback(img1, img3, flow=None, pyr_scale=0.5, levels=5, winsize=12,iterations= 2, poly_n=7, poly_sigma= 1.5, flags=0)
        magnitud1, angle1 = cv2.cartToPolar(flow1[..., 0], flow1[..., 1])
        magnitud2, angle2 = cv2.cartToPolar(flow2[..., 0], flow2[..., 1])

        # display dense optical flow
        # flow_x1 = flow1[1, :, :]
        # flow_y1 = flow1[0, :, :]
        flow_x1=flow1[..., 0]
        flow_y1=flow1[..., 1]

        print(flow_x1.shape, ": ",flow_y1.shape, ": ", img1.shape)

        flow_x1[img1==0]=np.nan
        flow_x1[img2==0]=np.nan
        flow_x1[img3==0]=np.nan
        flow_y1[img1==0]=np.nan
        flow_y1[img2==0]=np.nan
        flow_y1[img3==0]=np.nan

    
        flow_x2=flow2[..., 0]
        flow_y2=flow2[..., 1]
        # flow_x2 = flow2[1, :, :]
        # flow_y2 = flow2[0, :, :]

        flow_x2[img1==0]=np.nan
        flow_x2[img2==0]=np.nan
        flow_x2[img3==0]=np.nan
        flow_y2[img1==0]=np.nan
        flow_y2[img2==0]=np.nan
        flow_y2[img3==0]=np.nan

        ####Coh Thresh
        flow_x1[S12<Coh_Thresh]=np.nan
        flow_y1[S12<Coh_Thresh]=np.nan

        flow_x1[S13<Coh_Thresh]=np.nan
        flow_y1[S13<Coh_Thresh]=np.nan
        
        flow_x2[S12<Coh_Thresh]=np.nan
        flow_y2[S12<Coh_Thresh]=np.nan

        flow_x2[S13<Coh_Thresh]=np.nan
        flow_y2[S13<Coh_Thresh]=np.nan

        ####3  ####Vel thresh

        def vel_threshold(a, thresh= 0.063):
            h=a.shape[0]
            w=a.shape[1]
            print(a.shape)
            for k in range(h):
                for q in range(w):
                    if a[k,q] > thresh:
                        a[k,q]=np.nan
                    elif a[k,q] < -thresh:
                        a[k,q]=np.nan
            return a

        # flow_x1=vel_threshold(flow_x1, thresh=vel_thresh)
        # flow_x2=vel_threshold(flow_x2, thresh=vel_thresh)

        # flow_y1=vel_threshold(flow_y1, thresh=vel_thresh)
        # flow_x2=vel_threshold(flow_x2, thresh=vel_thresh)


        #############
        # flow_x1[flow_y1==np.nan]=np.nan
        # flow_y1[flow_x1==np.nan]=np.nan
        # flow_x2[flow_y2==np.nan]=np.nan
        # flow_y2[flow_x2==np.nan]=np.nan


        #################

        # from scipy import ndimage
        # flow_x1=ndimage.median_filter(flow_x1, size=20)
        # flow_x2=ndimage.median_filter(flow_x2, size=20)
        # flow_y1=ndimage.median_filter(flow_y1, size=20)
        # flow_y2=ndimage.median_filter(flow_y2, size=20)

        # flow_x1=cv2.bilateralFilter(flow_x1, 9, 11,11)
        # flow_x2=cv2.bilateralFilter(flow_x2, 9, 11,11)
        # flow_y1=cv2.bilateralFilter(flow_y1, 9, 11,11)
        # flow_y2=cv2.bilateralFilter(flow_y2, 9, 11,11)
    
        flow_x=[flow_x1,flow_x2]
        flow_x = np.stack(flow_x)
        

        flow_y=[flow_y1,flow_y2]
        flow_y=np.stack(flow_y)

    

        # flow_x=np.ma.array(flow_x, mask=(flow_x==0))
        # flow_y=np.ma.array(flow_y, mask=(flow_y==0))
        # std_x=np.nanstd(flow_x, axis=0)
        # std_x[std_x > 1]=np.nan
        # std_y=np.nanstd(flow_y, axis=0)
        # std_y[std_y > 1]=np.nan

        # mask_x=np.isnan(std_x)
        # mask_y=np.isnan(std_y)

        flow_x= np.mean(flow_x, axis=0)
        flow_y= np.mean(flow_y, axis=0)
        flow_y=flow_y*-1
        if Median_Filter==True:
            from scipy import ndimage
            flow_x=ndimage.median_filter(flow_x, size=20)
            flow_y=ndimage.median_filter(flow_y, size=20)
    

        flow_y=vel_threshold(flow_y, thresh=vel_thresh)
        flow_x=vel_threshold(flow_x, thresh=vel_thresh)
        
        maskx=np.isnan(flow_x)
        masky=np.isnan(flow_y)
        flow_x[masky]=np.nan
        flow_y[maskx]=np.nan

    

        
        # Taking a matrix of size 3 as the kernel
        # kernel = np.ones((3,3), np.float32)

        # flow_x = cv2.morphologyEx(flow_x, cv2.MORPH_OPEN, kernel)
        # flow_y = cv2.morphologyEx(flow_y, cv2.MORPH_OPEN, kernel)
        # flow_x = cv2.morphologyEx(flow_x, cv2.MORPH_CLOSE, kernel)
        # flow_y = cv2.morphologyEx(flow_y, cv2.MORPH_CLOSE, kernel)

        

        ##############

        flow_x[img1==0]=np.nan
        flow_x[img2==0]=np.nan
        flow_x[img3==0]=np.nan
        flow_y[img1==0]=np.nan
        flow_y[img2==0]=np.nan
        flow_y[img3==0]=np.nan

        ##Coh Thresh
        flow_x[S12<Coh_Thresh]=np.nan
        flow_y[S12<Coh_Thresh]=np.nan
        flow_x[S13<Coh_Thresh]=np.nan
        flow_y[S13<Coh_Thresh]=np.nan

        # number_imageframes=3
        # flow_x=flow_x/number_imageframes
        # flow_y=flow_y/number_imageframes

        ###

        # flow_x=flow_x/Delta_DD  
        # flow_y=flow_y/Delta_DD
        #Save Flowx and Flow y rasters
        filename=(str(Date1_YYYY+Date1_MM+Date1_DD) + "-" +
        str(Date2_YYYY+Date2_MM+Date2_DD)+ "-"+ str(Date3_YYYY+Date3_MM+Date3_DD)+".tif")
        
        print ("Now Processing Triplet Dates: ", filename[:-4])
        
        #Read DEM file
        pathhr=Path_to_DEMFile
        with rasterio.open(pathhr, 'r') as r1:
            demfile = r1.read(1)
            meta = r1.meta
        
        flow_xgr=flow_x /3/ int(Delta_DD) *3125.0 #convert pixel to cm(mm) and divide by difference of number of days between frame1 and frame3
        flow_ygr=flow_y /3/ int(Delta_DD) *3125.0

    

        

        mag_map1 = np.hypot(flow_xgr, flow_ygr)  # magnitude
        if udm_mask_option==True:
            mag_map1[udm_img1==0]=np.nan
            mag_map1[udm_img2==0]=np.nan
            mag_map1[udm_img3==0]=np.nan

        mag_map1[img1==0]=np.nan
        mag_map1[img2==0]=np.nan
        mag_map1[img3==0]=np.nan
        mag_map1[S12<Coh_Thresh]=np.nan
        mag_map1[S13<Coh_Thresh]=np.nan
        

        flow_ygr[mag_map1 == np.nan]=np.nan
        flow_xgr[mag_map1 == np.nan]=np.nan

        flow_ygr[flow_x==np.nan]=np.nan
        flow_xgr[flow_y==np.nan]=np.nan


        # mag_map1 = mag_map1[~is_outlier(mag_map1)]
        # flow_xgr = flow_xgr[~is_outlier(flow_xgr)]
        # flow_ygr = flow_ygr[~is_outlier(flow_ygr)]

        # flow_ygr[flow_ygr == 0]=np.nan
        # flow_xgr[flow_xgr == 0]=np.nan
        # mag_map1[mag_map1 == 0]=np.nan

        
    

        def interpolate_missing_pixels(
                image: np.ndarray,
                mask: np.ndarray,
                method: str = 'nearest',
                fill_value: int = 0):
            """
            :param image: a 2D image
            :param mask: a 2D boolean image, True indicates missing values
            :param method: interpolation method, one of
                'nearest', 'linear', 'cubic'.
            :param fill_value: which value to use for filling up data outside the
                convex hull of known pixel values.
                Default is 0, Has no effect for 'nearest'.
            :return: the image with missing values interpolated
            """
            

            h, w = image.shape[:2]
            xx, yy = np.meshgrid(np.arange(w), np.arange(h))

            known_x = xx[~mask]
            known_y = yy[~mask]
            known_v = image[~mask]
            missing_x = xx[mask]
            missing_y = yy[mask]


            #######

            xvalues = np.linspace(int(1), int(1+w), w) 
            yvalues = np.linspace(int(1), int(1+h), h)

            missing_x, missing_y = np.meshgrid(xvalues, yvalues)
            



            interp_values = interpolate.griddata(
                (known_x, known_y), known_v, (missing_x, missing_y),
                method=method, fill_value=fill_value
            )

            # interp_image = image.copy()
            # interp_image[missing_y, missing_x] = interp_values

            interp_image=interp_values

            return interp_image

        ##Reapply bad mask

        mag_map1[flow_xgr==np.nan]=np.nan

        flow_xgr[img1==0]=np.nan
        flow_ygr[img1==0]=np.nan
        mag_map1[img1==0]=np.nan

        if udm_mask_option==True:

            flow_xgr[udm_img1==0]=np.nan
            flow_ygr[udm_img1==0]=np.nan
            mag_map1[udm_img1==0]=np.nan

            flow_xgr[udm_img2==0]=np.nan
            flow_ygr[udm_img2==0]=np.nan
            mag_map1[udm_img2==0]=np.nan


            flow_xgr[udm_img3==0]=np.nan
            flow_ygr[udm_img3==0]=np.nan
            mag_map1[udm_img3==0]=np.nan


        with rasterio.open(output_VEL_Triplets + "/" + str(filename), 'w', **meta) as dst:
            dst.write(mag_map1, indexes=1)

        with rasterio.open(output_dirflowxn + "/" + str(filename), 'w', **meta) as dst:
            dst.write(flow_xgr, indexes=1)
        with rasterio.open(output_dirflowyn + "/" + str(filename), 'w', **meta) as dst:
            dst.write(flow_ygr, indexes=1)

###########################################################################Interpolation of data below
        ################
        # xgrmask=np.isnan(flow_x)
        # ygrmask=np.isnan(flow_y)
        # mag_mask=np.isnan(mag_map1)

        
        # flow_xgr=interpolate_missing_pixels(flow_xgr, xgrmask , "cubic")
        # flow_ygr_int=interpolate_missing_pixels(flow_ygr, ygrmask , "cubic")

        # flow_ygr[mag_map1 == np.nan]=np.nan
        # flow_xgr[mag_map1 == np.nan]=np.nan

        # mag_map1=interpolate_missing_pixels(mag_map1, ygrmask , "cubic")

        # ###Apply velocity threshold
        # flow_xgr_int=vel_threshold(flow_xgr_int, thresh=vel_thresh * 3125.0)
        # flow_ygr_int=vel_threshold(flow_ygr_int, thresh=vel_thresh * 3125.0)
        # mag_map1_int=vel_threshold(mag_map1_int, thresh=vel_thresh * 3125.0)

        #  ##Reapply bad mask

        # mag_map1[flow_xgr==np.nan]=np.nan

        # flow_xgr[img1==0]=np.nan
        # flow_ygr[img1==0]=np.nan
        # mag_map1[img1==0]=np.nan

        # flow_xgr[udm_img1==0]=np.nan
        # flow_ygr[udm_img1==0]=np.nan
        # mag_map1[udm_img1==0]=np.nan

        # flow_xgr[udm_img2==0]=np.nan
        # flow_ygr[udm_img2==0]=np.nan
        # mag_map1[udm_img2==0]=np.nan


        # flow_xgr[udm_img3==0]=np.nan
        # flow_ygr[udm_img3==0]=np.nan
        # mag_map1[udm_img3==0]=np.nan


        # ######Interpolation iter2

        #  ################
        # xgrmask=np.isnan(flow_x)
        # ygrmask=np.isnan(flow_y)
        # mag_mask=np.isnan(mag_map1)
        
        # flow_xgr_int=interpolate_missing_pixels(flow_xgr_int, xgrmask , "nearest")
        # flow_ygr_int=interpolate_missing_pixels(flow_ygr_int, ygrmask , "nearest")

        # flow_ygr_int[mag_map1 == np.nan]=np.nan
        # flow_xgr_int[mag_map1 == np.nan]=np.nan

        # mag_map1_int=interpolate_missing_pixels(mag_map1_int, ygrmask , "nearest")

        # flow_xgr_int=vel_threshold(flow_xgr_int, thresh=vel_thresh * 3125.0)
        # flow_ygr_int=vel_threshold(flow_ygr_int, thresh=vel_thresh * 3125.0)
        # mag_map1_int=vel_threshold(mag_map1_int, thresh=vel_thresh * 3125.0)


        ################



        # mag_map1_int[flow_xgr==np.nan]=np.nan

        # flow_xgr_int[img1==0]=np.nan
        # flow_ygr_int[img1==0]=np.nan
        # mag_map1_int[img1==0]=np.nan

        # flow_xgr_int[udm_img1==0]=np.nan
        # flow_ygr_int[udm_img1==0]=np.nan
        # mag_map1_int[udm_img1==0]=np.nan

        # flow_xgr_int[udm_img2==0]=np.nan
        # flow_ygr_int[udm_img2==0]=np.nan
        # mag_map1_int[udm_img2==0]=np.nan


        # flow_xgr_int[udm_img3==0]=np.nan
        # flow_ygr_int[udm_img3==0]=np.nan
        # mag_map1_int[udm_img3==0]=np.nan


        


        # mag_map1 = mag_map1[~is_outlier(mag_map1)]
        # flow_xgr = flow_xgr[~is_outlier(flow_xgr)]
        # flow_ygr = flow_ygr[~is_outlier(flow_ygr)]

        # mag_map1=np.percentile(mag_map1, 80)
        # flow_xgr=np.percentile(flow_xgr, 80)
        # flow_ygr=np.percentile(flow_ygr, 80)

        #############################################
        xname=output_dirflowxn + "/" + str(filename)
        yname=output_dirflowyn + "/" + str(filename)
        vname=output_VEL_Triplets + "/" + str(filename)

    
    
        # ###Save interpolated rasters
        # with rasterio.open(xname_int, 'w', **meta) as dst:
        #     dst.write(flow_xgr_int, indexes=1)

        # with rasterio.open(yname_int, 'w', **meta) as dst:
        #     dst.write(flow_ygr_int, indexes=1)
        # with rasterio.open(vname_int, 'w', **meta) as dst:
        #     dst.write(mag_map1_int, indexes=1)


    

        # print("max", mag_map1.max())
        # print("min", mag_map1.min())
        # print("std_mag", np.std(mag_map1))
        # print("mean_mag", np.mean(mag_map1))
        # print("-----------------------------")
        # print("Dates: ", item1[:-4], item2[:-4],  item3[:-4])
        # # print('default Flow_x Min', flow_xgr.min())
        # # print('default Flow_y Min', flow_xgr.min())

        # print("---------------------------------")



        ###########Resample data to remove noise and fill gaps
        def resample(input_raster="", output_raster="" , xres=3.125 , yres=3.125):
            

            ds = gdal.Open(input_raster)

            # resample
            dsRes = gdal.Warp(output_raster, ds, xRes = xres, yRes = yres, 
                            resampleAlg = "bilinear")

            # visualize
            # array = dsRes.GetRasterBand(1).ReadAsArray()
            # plt.figure()
            # plt.imshow(array)
            # plt.colorbar()
            # plt.show()

            # close your datasets!
            dsRes =None
            return output_raster

        if plot_option=="resampled":
            if not os.path.exists(output_dirflowxn + "/resampled" ):
                os.makedirs(output_dirflowxn + "/resampled")

            if not os.path.exists(output_dirflowyn + "/resampled" ):
                os.makedirs(output_dirflowyn + "/resampled")

            if not os.path.exists(output_VEL_Triplets + "/resampled" ):
                os.makedirs(output_VEL_Triplets + "/resampled")


            xname_int=output_dirflowxn + "/resampled" + "/" + str(filename)
            yname_int=output_dirflowyn + "/resampled" + "/" + str(filename)
            vname_int=output_VEL_Triplets + "/resampled" + "/" + str(filename)
            ew=resample(xname, xname_int, xres=xres, yres=yres)
            ns=resample(yname, yname_int, xres=xres, yres=yres)
            _2dvel=resample(vname, vname_int, xres=xres, yres=yres)
        else:
            print(" User selected to ignore resampling raster images ")
        #####Plot Timeseries

        #Function to read dem and prepare it to create hillshade for basemap

        def raster2array(geotif_file):
            metadata = {}
            dataset = gdal.Open(geotif_file)
            metadata['array_rows'] = dataset.RasterYSize
            metadata['array_cols'] = dataset.RasterXSize
            metadata['bands'] = dataset.RasterCount
            metadata['driver'] = dataset.GetDriver().LongName
            metadata['projection'] = dataset.GetProjection()
            metadata['geotransform'] = dataset.GetGeoTransform()
            mapinfo = dataset.GetGeoTransform()
            metadata['pixelWidth'] = mapinfo[1]
            metadata['pixelHeight'] = mapinfo[5]
        #     metadata['xMin'] = mapinfo[0]
        #     metadata['yMax'] = mapinfo[3]
        #     metadata['xMax'] = mapinfo[0] + dataset.RasterXSize/mapinfo[1]
        #     metadata['yMin'] = mapinfo[3] + dataset.RasterYSize/mapinfo[5]  
            metadata['ext_dict'] = {}
            metadata['ext_dict']['xMin'] = mapinfo[0]
            metadata['ext_dict']['xMax'] = mapinfo[0] + dataset.RasterXSize/mapinfo[1]
            metadata['ext_dict']['yMin'] = mapinfo[3] + dataset.RasterYSize/mapinfo[5]
            metadata['ext_dict']['yMax'] = mapinfo[3]
            
            metadata['extent'] = (metadata['ext_dict']['xMin'],metadata['ext_dict']['xMax'],
                                metadata['ext_dict']['yMin'],metadata['ext_dict']['yMax'])
            
            if metadata['bands'] == 1:
                raster = dataset.GetRasterBand(1)
                metadata['noDataValue'] = raster.GetNoDataValue()
                metadata['scaleFactor'] = raster.GetScale()
                
                # band statistics
                metadata['bandstats'] = {} #make a nested dictionary to store band stats in same 
                stats = raster.GetStatistics(True,True)
                metadata['bandstats']['min'] = round(stats[0],2)
                metadata['bandstats']['max'] = round(stats[1],2)
                metadata['bandstats']['mean'] = round(stats[2],2)
                metadata['bandstats']['stdev'] = round(stats[3],2)
                
                array = dataset.GetRasterBand(1).ReadAsArray(0,0,metadata['array_cols'],metadata['array_rows']).astype('float32')
                array[array==(metadata['noDataValue'])]=np.nan
                # array = array/metadata['scaleFactor']
                return array, metadata
            
            elif metadata['bands'] > 1:
                print('More than one band ... fix function for case of multiple bands')

        def hillshade(array,azimuth,angle_altitude):
            azimuth = 360.0 - azimuth 
            x, y = np.gradient(array)
            slope = np.pi/2. - np.arctan(np.sqrt(x*x + y*y))
            aspect = np.arctan2(-x, y)
            azimuthrad = azimuth*np.pi/180.
            altituderad = angle_altitude*np.pi/180.
            shaded = np.sin(altituderad)*np.sin(slope) + np.cos(altituderad)*np.cos(slope)*np.cos((azimuthrad - np.pi/2.) - aspect)

            return 255*(shaded + 1)/2

        

        
        
        dtm_array, dtm_metadata = raster2array(Path_to_DEMFile)
        if plot_option=="origional":

            srcx, src_meta = raster2array(xname)
            srcy, src_meta = raster2array(yname)
            srcv, src_meta = raster2array(vname)
        elif plot_option=="resampled":
            srcx, src_meta = raster2array(xname_int)
            srcy, src_meta = raster2array(yname_int)
            srcv, src_meta = raster2array(vname_int)
        
        srcx = st.resize(srcx, dtm_array.shape, mode='constant')
        srcy = st.resize(srcy, dtm_array.shape, mode='constant')
        srcv = st.resize(srcv, dtm_array.shape, mode='constant')

        # srcx_int, src_meta = raster2array(xname_int)
        # srcy_int, src_meta = raster2array(yname_int)
        # srcv_int, src_meta = raster2array(vname_int)
        
        # Use hillshade function on a DTM Geotiff
        hs_array = hillshade(dtm_array,335,45)
        
        def plot(hillshade, velocity, filename="" , title="EW Velocity", outputfolder=r"" , Delta_DD=1, cmap=cmap, masked=False, cbar_unit="mm/day"):
            
            outputfolder='Figs_analysis/timeseriesFigs' + "/" + outputfolder
        
            if not os.path.exists(outputfolder ):
                os.makedirs(outputfolder)

            velocity=velocity.astype(np.int16)
            diff = mcolors.LinearSegmentedColormap.from_list("", ["blue","white","red"])
            max=velocity.max()
            min=velocity.min()
            #print("min: ", min, " max: ", max)
            
            if min < 0 and max >0 :
                if Set_fig_MinMax==True:
                    min_n=-50
                    max_n=50
                    min=min_n
                    max=max_n
                    offset = mcolors.TwoSlopeNorm(vmin=min,
                                vcenter=0., vmax=max)  
                else:
                    offset = mcolors.TwoSlopeNorm(vmin=min,
                                vcenter=0., vmax=max)
                    
            else  : 
                if Set_fig_MinMax==True:
                    min_n=0
                    max_n=100
                    min=min_n
                    max=max_n
                    offset=mcolors.Normalize(vmin=min, vmax=max)
                else:
                    offset=mcolors.Normalize(vmin=min, vmax=max)

            # elif max <= 0 :
            #     offset=mcolors.Normalize()
                
            #scale = {"dx" : pixel_resolution_meter, "units" : "m"}
            fig, ax=plt.subplots(figsize=(10,5), nrows=1, ncols=1)

            axins = inset_axes(
            ax,
            width="5%",  # width: 3% of parent_bbox width
            height="75%",  # height: 50%
            loc="lower left",
            bbox_to_anchor=(1.05, 0., 1, 1),
            bbox_transform=ax.transAxes,
            borderpad=0)
            
            ax0=sb.imshow(hillshade, ax=ax,  cmap="gray", cbar=False, dx=3.125, units='m', color="k")

            if masked==True:
                velocity=velocity.astype("float32")
                velocity[velocity==0]=np.nan
            ax1=ax.imshow(velocity,  alpha=0.75, norm=offset, cmap=cmap )
            ax.set_title(str(title))
            # plt.colorbar(ax1, orientation="horizontal")
            a=fig.colorbar(ax1, cax=axins, extend="both")
            a.set_label(cbar_unit, labelpad=2, y=0.5, rotation=90)
            # title
            
            #set the subtitle
            dates_title="Dates:" +str(item1[:-4]) + "-"+ str(item2[:-4]) + "-"+ str(item3[:-4])
        
            ####
            # ax1=sb.imgplot(hillshade, ax=ax[1],  cmap="gray", cbar=False)
            # ax1=sb.imgplot(velocity_interpolated, ax=ax[1], alpha=0.75, cbar_label="mm/"+"day", cmap="gist_rainbow", **scale, showticks=True, orientation="h")
            plt.grid('on') 
            #ax[0].ticklabel_format(useOffset=True, style='plain') #do not use scientific notation #
            #ax[1].set_title("Interpolated: " + str(title) + str(filename[:-4]))
            # scalebar = ScaleBar(3.125, color="k") # 1 pixel = 3.125 meter
            # plt.gca().add_artist(scalebar)
            figname=filename[:-4]+".jpg"
            
            bn=dtm_array.shape
            xs=bn[1]
            ys=bn[0]
            ax.set_xlim(0, xs)
            ax.set_ylim(ys,0)
            ax.text(0.5, - 0.07, dates_title,horizontalalignment='center', verticalalignment='center', transform=ax.transAxes,
            bbox=dict(facecolor='red', alpha=0.5))   


            plt.savefig(outputfolder + "/" + figname, dpi=300)

            if show_figure==True:
                plt.show()
            else:
                plt.close(fig)

        _1=plot(hs_array, srcx, filename, title="EW Mean-Velocity", outputfolder="EW", Delta_DD= int(Delta_DD), cmap=cmap , masked=True, cbar_unit="mm/day" )
        _2=plot(hs_array, srcy, filename, title="NS Mean-Velocity", outputfolder="NS" , Delta_DD= int(Delta_DD), cmap=cmap, masked=True, cbar_unit="mm/day")
        _3=plot(hs_array, srcv, filename, title="2D Mean-Velocity", outputfolder="2D" , Delta_DD= int(Delta_DD),  cmap=cmap, masked=True, cbar_unit="mm/day")

    

        S12=exposure.rescale_intensity(S12, out_range=(0, 100)).astype(np.float32)
        S13=exposure.rescale_intensity(S13, out_range=(0, 100)).astype(np.float32)
        S12[S12==100]=np.nan
        S13[S13==100]=np.nan
        S12[S12<Coh_Thresh*100]=np.nan
        S13[S13<Coh_Thresh*100]=np.nan
        
        _4=plot(hs_array, S12, filename, title="Similarity Index Between Dates 1&2", outputfolder="S12" , Delta_DD= int(Delta_DD) , cmap=cmap, masked=True, cbar_unit="Similarity-Index %")
        _5=plot(hs_array, S13, filename, title="Similarity Index Between Dates 1&3", outputfolder="S13" , Delta_DD= int(Delta_DD), cmap=cmap, masked=True, cbar_unit="Similarity-Index %")


        # fig, ax = plt.subplots()
        # img = ax.imshow(S13)
        # fig.colorbar(img, ax=ax) 
        # plt.show()


        f=f+1
        
        # plt.imshow(img3, cmap="gray")
        # plt.colorbar(plt.imshow(mag_map1, cmap="gist_rainbow", alpha=0.7))
        # plt.title(Delta_DD)
        #plt.show()
        
        if (len(img_list) - 2) == f:
            break
        print("Finished Processing Triplet Dates: ", filename[:-4])
        os.system("printf '\033c'")
        
        
        
        
        
    print (" process is compeleted")
   
        







def unzip(zipdir, dst_dir ):

    """
    This program unzips all the zip products into one folder

    Parameters
    ----------
   

    zipdir : str
        path to directory contains all the zipfiles

    dst_dir : str
        path to destination folder to copy all unzipped products.

    Returns
    -------
    unzip folder

    """
    import os 
    import shutil
    from zipfile import ZipFile
    import glob
    import os
    from os import listdir
    from os.path import isfile, join
    if not os.path.exists( dst_dir):
        os.makedirs( dst_dir)
    dst_dir=dst_dir
    zipdir=zipdir
    zip_list = [f for f in sorted(os.listdir(zipdir)) if isfile(join(zipdir, f))]

    for n in range (0, len(zip_list)):
        with ZipFile(join(zipdir,zip_list[n]), 'r') as zipObject:
            listOfFileNames = zipObject.namelist()
            for fileName in listOfFileNames:
                if fileName.endswith('.tif'):
                    # Extract a single file from zip
                    zipObject.extract(fileName, dst_dir)
                    print('All the tif files are extracted')


def copyImage_Data(path_to_unzipped_folders=r"", Path_to_raster_tifs=""):
    """
    This program copy all the raster images.

    Parameters
    ----------

    path_to_unzipped_folders : str

    
    Path_to_raster_tifs : str

    Returns
    -------
    rasters

    """
    import os 
    import shutil
    from zipfile import ZipFile
    import glob
    import os
    from os import listdir
    from os.path import isfile, join

    if not os.path.exists( Path_to_raster_tifs):
        os.makedirs( Path_to_raster_tifs)

    S2_path = os.path.join (path_to_unzipped_folders)

    for subdir, dirs, files in os.walk(S2_path):
        for file in files:
            filepath = subdir + os.sep + file
            if filepath.endswith("BGRN_SR_clip.tif"):
                print (filepath)
                shutil.copy2(filepath, Path_to_raster_tifs)


def copyUDM2_Mask_Data(path_to_unzipped_folders=r"", Path_to_UDM2raster_tifs=r""):

    """
    This program copy all  raster masks.

    Parameters
    ----------

    path_to_unzipped_folders : str
        file extension must end with udm2_clip.tif

    Path_to_UDM2raster_tifs : str

    Returns
    -------
    rasters

    """
    
    import os 
    import shutil
    from zipfile import ZipFile
    import glob
    import os
    from os import listdir
    from os.path import isfile, join

    if not os.path.exists( Path_to_UDM2raster_tifs):
        os.makedirs( Path_to_UDM2raster_tifs)

    S2_path = os.path.join (path_to_unzipped_folders)

    for subdir, dirs, files in os.walk(S2_path):
        for file in files:
            filepath = subdir + os.sep + file
            if filepath.endswith("udm2_clip.tif"):
                print (filepath)
                shutil.copy2(filepath, Path_to_UDM2raster_tifs)

    



def Time_Series(stacked_raster_EW=r"", stacked_raster_NS=r"", velocity_points=r"", dates_name=r"", output_folder="", outputFilename="",
                rasteriz_mean_products=True, std=1, VEL_Scale='year' , velocity_mode="mean", master_reference=False):
    
    '''
    This program uses candiate velocity points from stackprep function and performs linear interpolation in time-domain to calibrate
    stacked velocity. Additionally produces corrected timeseries velocity(daily) in a shapefile.
    
    Parameters
    ----------
    
    stacked_raster_EW : str
    
    stacked_raster_NS : str
    
    velocity_points : str 
        Velcity Candidate points
    
    dates_name : str
        text file include name of each date in format YYYYMMDD
    
    output_folder : str
    
    outputFilename : str
    
    VEL_Scale   : str
        'year' , "month" or empty  to calculate velocity within provided dataset date range
    
    velocity_mode   : str
        "mean" or "linear"
        
    master_reference : bool
        True if calculate TS to a single reference date, False if calculate TS to subsequent Reference dates
    
    Returns
    -------
    
    Time-series shape file of velocity and direction EW, NS, and 2D(resultant Velocity and direction)
    
    '''
    import rasterio
    import os
    from os.path import isfile, join
    import numpy as np
    import glob
    import geowombat as gw
    import pandas as pd
    import geopandas as gpd
    import scipy.stats as stats
    from geocube.api.core import make_geocube
    from geocube.rasterize import rasterize_points_griddata, rasterize_points_radial
    from functools import partial
    from datetime import datetime
    from dateutil import parser
    
    def Helper_Time_Series(stacked_raster=r"", velocity_points=r"", dates_name=r"", output_folder="", outputFilename="", std=1 , VEL_Scale=VEL_Scale):
        
        '''
        stacked_raster: Path to raster stack .tif file
        
        velocity_points: Path to velocity points in arcmap shapfile format .shp
        
        dates_name: path to text file contains date names of each time series triplets .txt
        
        output_folder: Path to output Folder
        
        outputFilename: name for out time-series velocity shapefile
        '''
        
        import os
        import numpy as np
        import glob
        import geowombat as gw
        import pandas as pd
        import geopandas as gpd
        from sklearn.linear_model import LinearRegression
        
        if not os.path.exists(output_folder):
                os.makedirs(output_folder)
    
        
        #Open Raster stack, extract pixel info into shape file

        with gw.open(stacked_raster, stack_dim='time') as src:
            print(src)
            df = src.gw.extract(velocity_points)

        #Import names to label timeseries data    
        names = []
        dnames=[]
        with open(dates_name, 'r') as fp:
            for line in fp:
                # remove linebreak from a current name
                # linebreak is the last character of each line
                x = 'D'+ line[:-1]

                # add current item to the list
                names.append(x)
                dnames.append(x[:-18])

        print (len(dnames))
        print(len(df.columns))

        cci=(len(df.columns)- len(dnames))
        df2=df.iloc[:, cci:]
        cc=np.arange(1,cci)
        #Add Timesereises column names
        
        #find outliers using z-score iter 1
        lim = np.abs((df2[cc] - df2[cc].mean(axis=1)) / df2[cc].std(ddof=0, axis=1)) < std
        
        # # # replace outliers with nan
        df2[cc]= df2[cc].where(lim, np.nan)
        
        
        
        df2[cc] = df2[cc].astype(float).apply(lambda x: x.interpolate(method='linear', limit_direction='both'), axis=1).ffill().bfill()
       
        
        df2=df2.T
        
        #find outliers using z-score iter 2
        lim = np.abs((df2 - df2.mean(axis=0)) / df2.std(ddof=0,axis=0)) < std
        #lim=df2.apply(stats.zscore, axis=1) <1
        # # # replace outliers with nan
        df2= df2.where(lim, np.nan)
        
        df2= df2.astype(float).apply(lambda x: x.interpolate(method='linear', limit_direction='both'), axis=0).ffill().bfill()
        
        for col in df2.columns:
            #print (col)
            #df2[col]=pd.to_numeric(df2[col])
            df2[col]= df2[col].interpolate(method='index', axis=0).ffill().bfill()
        
        df2=df2.T
            
           
        #Add Timesereises column names
        df2.columns = dnames
        
        #Calculate Linear Velocity for each data point
        def linear_VEL(df, dnames):
            from datetime import datetime 
            from scipy import stats
            # def best_fit_slope_and_intercept(xs,ys):
            #     from statistics import mean
            #     xs = np.array(xs, dtype=np.float64)
            #     ys = np.array(ys, dtype=np.float64)
            #     m = (((mean(xs)*mean(ys)) - mean(xs*ys)) /
            #         ((mean(xs)*mean(xs)) - mean(xs*xs)))
                
            #     b = mean(ys) - m*mean(xs)
                
            #     return m, b
            dd_list=[x.replace("D", "") for x in dnames]
            dates_list=([datetime.strptime(x, '%Y%m%d') for x in dd_list])
            days_num=[( ((x) - (pd.Timestamp(year=x.year, month=1, day=1))).days + 1) for x in dates_list]
            days_num=list(range(0, len(dnames)))
            dslope=[]
            std_slope=[]
            for index, dr in df.iterrows():
                #if index==0:
                rows=df.loc[index, :].values.flatten().tolist()
                row_values=rows
                # dfr = pd.DataFrame(dr).transpose()
                # dfr = dfr.loc[:, ~dfr.columns.str.contains('^Unnamed')]
            
                #slopeVEL=best_fit_slope_and_intercept(days_num, row_values)
                #print("slope", slopeVEL[0])
                slope, intercept, r_value, p_value, std_err = stats.linregress(days_num, row_values)
                dslope.append(slope)
                std_slope.append(std_err)
            return dslope, std_slope
        
        
        
        
            
        
        ###########################################################################
  
        
        dnames_new=[x.replace("D", "") for x in dnames]
        def input_dates(start_date="YYYYMMDD", end_date="YYYYMMDD"):
            start_date1=parser.parse(start_date)
            end_date2=parser.parse(end_date)
            date_list_start=[]
            date_list_end=[]
            for idx, item in enumerate(dnames_new):
                #filepath1, img_name = os.path.split(item) 
                str_date1=item
                str_date2=dnames_new[len(dnames_new)-1]
                #input start date
                date_time1 = parser.parse(str_date1)
                date_list_start.append(date_time1)
                #input end date
                date_time2 = parser.parse(str_date2)
                date_list_end.append(date_time2)
            st_date=min(date_list_start, key=lambda d: abs(d - start_date1))
            text_date1=st_date.strftime("%Y%m%d")
            End_date=min(date_list_end, key=lambda d: abs(d - end_date2))
            No_ofDays=(End_date-st_date).days
            
            text_date2=End_date.strftime("%Y%m%d")
            return [text_date1, text_date2, No_ofDays]

        velocity_scale=(input_dates(start_date=dnames_new[0], end_date=dnames_new[len(dnames_new)-1]))
        
        #################################
        # for idx, row in df2[dnames].iterrows():
        #     lim = np.abs((row[dnames] - df2[dnames]()) / row[dnames].std(ddof=0)) < 1
        #     row[dnames]= row[dnames].where(lim, np.nan)
        #     row[dnames] = row[dnames].astype(float).apply(lambda x: x.interpolate(method='linear', limit_direction='both'), axis=1).ffill().bfill()
            
        
        print (df2.describe())
        temp_df=pd.DataFrame()
        temp_df[dnames[0]]=df2[dnames[0]]
        #Choosing first date as reference for Time Series
        
        if master_reference==True:
            
            df2 = df2.sub(df2[dnames[0]], axis=0)
        else:
            
            df2=df2.diff(axis = 1, periods = 1)
        # count=0
        # for idx, col in enumerate(df2.columns):
        #     df2[col] = df2[col].sub(df2[dnames[count]], axis=0)
        #     count=count+1
            
       
        df2[dnames[0]]=0
            
        linear_velocity=linear_VEL(df2[dnames], dnames)
        out=df2
        if velocity_mode=="mean":
            out['VEL']=out[dnames].mean(axis=1)
            out['VEL_STD']=out[dnames].std(axis=1)
        elif velocity_mode=="linear":
            out['VEL']=linear_velocity[0]
            out['VEL_STD']=linear_velocity[1]
        if VEL_Scale=="month": 
            out['VEL']=out['VEL']/velocity_scale[2] * 30
            out['VEL_STD']=out['VEL_STD']/velocity_scale[2] *30
        elif VEL_Scale=="year":
            out['VEL']=out['VEL']/velocity_scale[2] * 365
            out['VEL_STD']=out['VEL_STD']/velocity_scale[2] * 365
            
        
            
        out['geometry']=df['geometry']
        out['CODE']=df['SiteID']
        #out[dnames[0]]=temp_df[dnames[0]]
        # out['HEIGHT']=0
        # out['H_STDEV']=0
        #out['V_STDEV']=out[dnames].std(axis=1)
        #out['COHERENCE']=0
        #out['H_STDEF']=0
        out['x']=df['x']
        out['y']=df['y']

        col_titles=['CODE','geometry','x', 'y', 'VEL', 'VEL_STD' ]+dnames
        out = out.reindex(columns=col_titles)
        
        

        geo_out=gpd.GeoDataFrame(out, geometry='geometry', crs=df.crs)

        geo_out.to_file(output_folder +"/" + outputFilename)
        (geo_out)

        return geo_out, dnames, linear_VEL
    
    EW=Helper_Time_Series(stacked_raster='raster_stackxn.tif', velocity_points=velocity_points ,
                             dates_name='Names.txt', output_folder='stack_data/TS', outputFilename="TS_EW_"+ os.path.basename(velocity_points), std=std, VEL_Scale=VEL_Scale)
                             
    NS=Helper_Time_Series(stacked_raster='raster_stackyn.tif', velocity_points=velocity_points, 
                             dates_name='Names.txt', output_folder='stack_data/TS', outputFilename="TS_NS_"+ os.path.basename(velocity_points), std=std, VEL_Scale=VEL_Scale)
    
    
    if outputFilename=="":
            outputFilename= "TS_2D_"+ os.path.basename(velocity_points)
            
            
    gdf_ew=EW[0]
    gdf_ns=NS[0]
    dnames=NS[1]
    df_2D_VEL=pd.DataFrame()
    df_2D_VEL['CODE']=gdf_ew['CODE']
    df_2D_VEL['geometry']=gdf_ew['geometry']
    df_2D_VEL['x']=gdf_ew['x']
    df_2D_VEL['y']=gdf_ew['y']
    
   
   #Calculate resultant velocity magnitude
    for col in gdf_ew[dnames].columns:
       
        df_2D_VEL[col]=np.hypot(gdf_ns[col],gdf_ew[col])
       
    df_2D_VEL['VEL_MEAN']=df_2D_VEL[dnames].mean(axis=1)
    df_2D_VEL['V_STDEV']=df_2D_VEL[dnames].std(axis=1)
    #we call linear velocity function from above then reuse it below to replace VEL_2D Mean an STD below for lines
    # linear_2D_Velocity_function=EW[2]
    # linear_2D_Velocity=linear_2D_Velocity_function(df_2D_VEL[dnames], dnames)
    # df_2D_VEL['VEL']=linear_2D_Velocity[0]
    # df_2D_VEL['V_STDEV']=linear_2D_Velocity[1]
    #############################
    col_titles=['CODE','geometry','x', 'y', 'VEL_MEAN' , 'V_STDEV' ]+ dnames 
    df_2D_VEL = df_2D_VEL.reindex(columns=col_titles)
    gdf_2D_VEL=gpd.GeoDataFrame(df_2D_VEL, geometry='geometry', crs=gdf_ew.crs)
    
    
    
    gdf_2D_VEL.to_file(output_folder +"/" + outputFilename)
    
    
    #Calculate resultant velocity direction
    
    dir_df_2D_VEL=pd.DataFrame()
    dir_df_2D_VEL['CODE']=gdf_ew['CODE']
    dir_df_2D_VEL['geometry']=gdf_ew['geometry']
    dir_df_2D_VEL['x']=gdf_ew['x']
    dir_df_2D_VEL['y']=gdf_ew['y']
    
    newcol_dir_list=[]
    for col in gdf_ew[dnames].columns:
        newcol_dir= col
        newcol_dir_list.append(newcol_dir)
        dir_df_2D_VEL[newcol_dir]=np.arctan2(gdf_ns[col],gdf_ew[col])
        dir_df_2D_VEL[newcol_dir]=np.degrees(dir_df_2D_VEL[newcol_dir])
        dir_df_2D_VEL[newcol_dir]=(450-dir_df_2D_VEL[newcol_dir]) % 360
    dir_df_2D_VEL['VELDir_MEAN']=dir_df_2D_VEL[newcol_dir_list].mean(axis=1)
    col_titles=['CODE','geometry','x', 'y', 'VELDir_MEAN'  ]+ newcol_dir_list
    dir_df_2D_VEL = dir_df_2D_VEL.reindex(columns=col_titles)
    dir_gdf_2D_VEL=gpd.GeoDataFrame(dir_df_2D_VEL, geometry='geometry', crs=gdf_ew.crs)
    
    dir_gdf_2D_VEL.to_file(output_folder +"/" + outputFilename[:-4]+"_dir.shp")
    
    #Calcuate Mean Corrected velocity products MEAN X, Y, 2D and Dir
    corrected_mean_products=pd.DataFrame()
    corrected_mean_products['CODE']=gdf_ew['CODE']
    corrected_mean_products['geometry']=gdf_ew['geometry']
    corrected_mean_products['x']=gdf_ew['x']
    corrected_mean_products['y']=gdf_ew['y']
    corrected_mean_products['VEL_E']=gdf_ew['VEL']
    corrected_mean_products['VEL_N']=gdf_ns['VEL']
    #corrected_mean_products['VEL_2D']=df_2D_VEL['VEL_MEAN']
    corrected_mean_products['VEL_2D']=df_2D_VEL['VEL_MEAN']
    corrected_mean_products['2DV_STDEV']=df_2D_VEL['V_STDEV']
    corrected_mean_products['VEL_2DDir']=dir_df_2D_VEL['VELDir_MEAN']
    corrected_mean_products_geo=gpd.GeoDataFrame(corrected_mean_products, geometry='geometry', crs=gdf_ew.crs)
    
    corrected_mean_products_geo.to_file(output_folder +"/" + outputFilename[:-4]+"_mean.shp")
    
    
    if rasteriz_mean_products==True:
        
        corrected_means_list=['VEL_E', 'VEL_N', 'VEL_2D', 'VEL_2DDir', '2DV_STDEV']
       
        for id, col_means in enumerate(corrected_means_list):
                
            geo_grid_vel = make_geocube(
                        vector_data=corrected_mean_products_geo,
                        measurements=[col_means],
                        resolution=(10, 10),
                    output_crs="epsg:32610",
                rasterize_function=partial(rasterize_points_radial, method="nearest") )
            
            geo_grid_vel[col_means].rio.to_raster("stack_data/"+ str(col_means)+'.tif')
            with rasterio.open("stack_data/"+ str(col_means)+'.tif', 'r+')as src:
                meta=src.meta
                src_data=src.read()
            # src.close() # close the rasterio dataset
            # os.remove("stack_data/"+ str(col_means)+'.tif') # delete the file 
            with rasterio.open("stack_data/"+ str(col_means)+'1.tif', 'w+', **meta) as dst:
                dst.write(src_data)
                               




def Filter_PreProcess(unfiltered_folderPath=r"", UDM2_maskfolderPath=r"", outpath_dir=r"" , Udm_Mask_Option=False):


    """
    This program prepare and uses filters to balanace raster image brightness

    Parameters
    ----------
    
    unfiltered_folderPath : str

    UDM2_maskfolderPath : str

    outpath_dir : str

    Udm_Mask_Option : bool
        False if True the program uses planetlabs imagery unusable pixel mask to ignore and mask bad image pixels 

    Returns
    -------
    geotif rasters
        Filtered geotif rasters
    Figures
        plotted filtered rasters and mask for bad pixels

    """
    
    import glob
    from itertools import count
    import os
    import rasterio
    from rasterio.plot import show
    from skimage import exposure
    import numpy as np
    import cv2
    from rasterio.plot import show_hist
    import matplotlib.pyplot as plt

    import seaborn_image as sea_img
    
    
    mypath = unfiltered_folderPath
    print("strat working on folder", mypath)

    #Setup Folder directories
    
    if not os.path.exists(outpath_dir):
        os.makedirs(outpath_dir)

    figs_dir=r"Figs_analysis/filter_figs"
    fig_masks=r"Figs_analysis/mask_figs"

    if not os.path.exists(figs_dir):
        os.makedirs(figs_dir)
    if not os.path.exists(fig_masks):
        os.makedirs(fig_masks)

    img_list=sorted(glob.glob(unfiltered_folderPath+ "/" +"*.tif"))
    udm_list=sorted(glob.glob(UDM2_maskfolderPath + "/" + "*.tif"))
    


    ###############

     
    
    
    for idx, item in enumerate(img_list):

        print(" start Processing Image Number: ", idx, ":", item)
        #img_item=img_list[idx]
        #msk_item=udm_list[idx]
        filepath1, filename = os.path.split(img_list[idx])
        filename=filename.replace("-", "")
        #rgb=rgb_normalize(img_list[idx])
        img_src=rasterio.open(img_list[idx])
        
        meta=img_src.meta
        # meta.update({'nodata':0})
        # meta.update({'count': 3})
        # meta.update({'dtype': np.uint8})

        

        if Udm_Mask_Option==True:
            udm_src=rasterio.open(udm_list[idx])


            clean_mask = udm_src.read(1).astype(bool)
            snow_mask = udm_src.read(2).astype(bool)
            shadow_mask = udm_src.read(3).astype(bool)
            light_haze_mask = udm_src.read(4).astype(bool)
            heavy_haze_mask = udm_src.read(5).astype(bool)
            cloud_mask = udm_src.read(6).astype(bool)
            bad_mask=snow_mask+shadow_mask+cloud_mask+light_haze_mask+heavy_haze_mask
            minimum_mask=shadow_mask+cloud_mask+snow_mask
            bad_mask=~clean_mask + snow_mask

        img_data = img_src.read([3, 2, 1], masked=True)/10000
        img_data_copy=img_data.copy()

        if Udm_Mask_Option==True:
            img_data.mask = bad_mask 
            img_data = img_data.filled(fill_value=0)
        
        def normalize(array):
            array_min, array_max = array.min(), array.max()
            (array - array_min) / (array_max - array_min)
            array=exposure.rescale_intensity(array, out_range=(0, 255)).astype(np.uint8)
            return array
        def pct_clip(array,pct=[2,98]):
            array_min, array_max = np.nanpercentile(array,pct[0]), np.nanpercentile(array,pct[1])
            clip = (array - array_min) / (array_max - array_min)
            clip[clip>1]=1
            clip[clip<0]=0
            return clip

        # Convert to numpy arrays
        r_o = img_src.read(3)
        b_o = img_src.read(2)
        g_o = img_src.read(1)

        r_o=exposure.rescale_intensity(r_o, out_range=(0, 255)).astype(np.uint8)

        b_o=exposure.rescale_intensity(b_o, out_range=(0, 255)).astype(np.uint8)

        g_o=exposure.rescale_intensity(g_o, out_range=(0, 255)).astype(np.uint8)

        rgb_img_data = np.dstack((r_o, b_o, g_o))
        rgb_img_data=np.transpose(rgb_img_data, (2,0,1))

        if UDM2_maskfolderPath!="":
            r = exposure.equalize_hist(r_o, mask=clean_mask+ ~snow_mask)
            b = exposure.equalize_hist(b_o, mask=clean_mask+ ~snow_mask)
            g = exposure.equalize_hist(g_o, mask=clean_mask+ ~snow_mask)
        else:
            r = exposure.equalize_hist(r_o)
            b = exposure.equalize_hist(b_o)
            g = exposure.equalize_hist(g_o)
            
        r = exposure.equalize_adapthist(r, kernel_size=128, clip_limit=0.01, nbins=256)
        b = exposure.equalize_adapthist(b, kernel_size=128, clip_limit=0.01, nbins=256)
        g = exposure.equalize_adapthist(g, kernel_size=128, clip_limit=0.01, nbins=256)

        r=r.astype("float64")
        b=b.astype("float64")
        g=g.astype("float64")

        r[r_o==0]=0
        b[b_o==0]=0
        g[g_o==0]=0

        r=pct_clip(r)
        b=pct_clip(b)
        g=pct_clip(b)

        rgb = np.dstack((r, b, g))
        rgb=np.transpose(rgb, (2,0,1))
        #rgb[rgb==0]=255
        if Udm_Mask_Option==True:
            rgb[img_data==0]=0
            
        else:
            rgb[img_data_copy==0]=0
            


        print (rgb.shape)
        
        rgb=exposure.rescale_intensity(rgb, out_range=(0, 255)).astype(np.uint8)
        
        with rasterio.open( outpath_dir + "/" + str(filename), "w", driver='GTiff', width=img_src.shape[1],
            height=img_src.shape[0], count=3, dtype='uint8', crs=img_src.crs, transform=img_src.transform, nodata=0) as dst:
            dst.write(rgb)
        
        ##Set plot extent
        # import rioxarray as rxr
        # from rasterio.plot import plotting_extent
        # rs = rxr.open_rasterio(udm_list[0], masked=True)
        #rs_plotting_extent = plotting_extent(rs[0], rs.rio.transform())
        fig1, ((ax12, ax22), (ax32, ax42) ) = plt.subplots(2, 2, figsize=(30,20))
        show(rgb_img_data, ax=ax12)
        show_hist(rgb_img_data, ax=ax22, bins=10, lw=0.0, 
        stacked=False, alpha=0.3, histtype='stepfilled', density=True, title="Initial: "+str(filename[:-4]))
        ax22.get_legend().remove()
        show(rgb, ax=ax32 )
        show_hist(rgb, ax=ax42, bins=10, lw=0.0, 
        stacked=False, alpha=0.3, histtype='stepfilled', density=True, title="CLAHE: "+str(filename[:-4]))
        ax42.get_legend().remove()
        bn=img_data.shape
        xs=bn[2]
        ys=bn[1]
        ax12.set_xlim(0, xs)
        ax12.set_ylim(ys,0)
        ax32.set_xlim(0, xs)
        ax32.set_ylim(ys,0)
        
        fig1.savefig(figs_dir+ "/" +  str(filename[:-4])+ ".jpg", dpi=150)
        plt.close(fig1)
        #plt.show()

        if UDM2_maskfolderPath!="":
            fig2, (ax12 , ax22) = plt.subplots(2,1, figsize=(15,10))
            show(rgb, ax=ax12)
            show(~clean_mask, cmap="binary", ax=ax22)
            #plt.show()
            bn=img_data.shape
            xs=bn[2]
            ys=bn[1]
            ax12.set_xlim(0, xs)
            ax12.set_ylim(ys,0)
            ax22.set_xlim(0, xs)
            ax22.set_ylim(ys,0)
            fig2.savefig(fig_masks + "/" +  str(filename[:-4])+ ".jpg", dpi=150)
            plt.close(fig2)
        if idx == len(img_list):
            break
        print("process is completed")
        os.system( 'cls' )
    print("process is completed")
        
        
        
        
        
        
        
        
        
        
        
        
        # #show(bad_mask, title="bad_mask: "+str(udm_list[idx]), cmap="binary")

        # if Color==True:
        #         RGB=[3,2,1]
        #         img_data = img_src.read([RGB[0], RGB[1], RGB[2]])/1000  # apply RGB ordering and scale down
        #         img_copy=img_data.copy()
        #         meta.update({'driver':'GTiff',
        #          'width':img_src.shape[1],
        #          'height':img_src.shape[0],
        #          'count':3,
        #          'dtype':'float32',
        #          'crs':img_src.crs, 
        #          'transform':img_src.transform,
        #          'nodata':0})
        #         if Udm_Mask_Option==True:
        #             img_data = img_src.read([RGB[0], RGB[1], RGB[2]], masked=True)/1000
        #             img_data.mask = bad_mask
        #             img_data = img_data.filled(fill_value=0)
        #         else:
        #             img_data=img_data

        # elif Color==False:
        #     clean_mask = udm_src.read(1).astype(bool)
        #     gray=2
        #     img_data = img_src.read(gray)  # apply RGB ordering and scale down
        #     img_data=img_data.astype("float")
        #     img_data[clean_mask==False]=np.nan
        #     img_copy=img_data.copy()
        #     meta.update({'driver':'GTiff',
        #          'width':img_src.shape[1],
        #          'height':img_src.shape[0],
        #          'count':1,
        #          'dtype':'float32',
        #          'crs':img_src.crs, 
        #          'transform':img_src.transform,
        #          'nodata':0})


        # print ("img_data shape: ", img_data.shape)

        # img_data=exposure.rescale_intensity(img_data, out_range=(0, 255)).astype(np.uint8)
        
        # #Global Histogram Equalization
        # img_histo_equ = exposure.equalize_hist(img_data, mask=clean_mask)

        
        # img_histo_equ[img_copy==0]=0

        # img_histo_equ=exposure.rescale_intensity(img_histo_equ, out_range=(0, 255)).astype(np.uint8)

        # #Apply CLAHE Filter 1 and 2
        # filteredimage = exposure.equalize_adapthist(img_histo_equ, clip_limit=0.01, nbins=256)

        # filteredimage[img_copy==0]=0

        # if Udm_Mask_Option==True:
        #     filteredimage=filteredimage.astype("float")
        #     filteredimage[clean_mask==False]=np.nan

        # filteredimage=exposure.rescale_intensity(filteredimage, out_range=(0, 255)).astype(np.uint8)

        # filteredimage = exposure.equalize_adapthist(filteredimage, clip_limit=0.01, nbins=256)

        # filteredimage[img_copy==0]=0

        # if Udm_Mask_Option==True:
        #     filteredimage=filteredimage.astype("float")
        #     filteredimage[clean_mask==False]=np.nan

        
        # filteredimage=exposure.rescale_intensity(filteredimage, out_range=(0, 255)).astype(np.uint8)
       
       
        
    
        # filteredimage[img_copy==0]=0
        # if Udm_Mask_Option==True:
        #     filteredimage=filteredimage.astype("float")
        #     filteredimage[clean_mask==False]=np.nan

        # #filteredimage=np.transpose(filteredimage, (1,0))
        # filteredimage=np.transpose(filteredimage, (0,1, 2))
        # print (filteredimage.shape)
        # with rasterio.open(outpath_dir+ "/" + str(filename),  'w+', **meta) as dst:
        #     dst.write(filteredimage)


        


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['akhdefo_functions']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'akhdefo-functions',
    'version': '2.2.38',
    'description': 'Land Deformation Monitoring Using Optical Satellite Imagery',
    'long_description': "\n# Akhdefo: \n# Background of Akh-Defo:\nAKh-Defo is combination of two different words 1) Akh in Kurdish language means land, earth or soil (origion of the word is from Kurdish badini dailect) 2) Defo is short of English word deformation.\n\n# Recommended Citation:\nMuhammad M, Williams-Jones G, Stead D, Tortini R, Falorni G and Donati D (2022) Applications of ImageBased Computer Vision for Remote Surveillance of Slope Instability.Front. Earth Sci. 10:909078. doi: 10.3389/feart.2022.909078\n\n# Updates:\n* Akhdefo version one is deprecated please use Akhdefo version 2.\n* Akhdefo now can run on the cloud for real-time processing\n* Akhdefo now consist of 12 Modules that performs end to end python-based GIS and Image Processing and Custumized Figure generation\n\n# Usage:\n* [download anaconda environment file](akhdefov2.yml) \n\nPlease visit the GITHUB homepage to download the file\n\n* [download python package requirement text](pip_req.txt) \n\nPlease visit the GITHUB homepage to download the file\n\n* Create new python Anaconda environment using the below command\n\n```python\nconda env create -f akhdefov2.yml\n\n```\n\n* Install required python packages using below command\n\n```python\npip install -r pip_req.txt\n```\n\n* Now install Akhdefo using the below command\n\n```python\npip install akhdefo-functions\n```\n\n# User Guide\n* Under construction will be released soon!\n\n# Akhdefo Functions Summary\n\n##  Preprocessing functions to unzip and sort images\n\n```python\nfrom akhdefo_functions import copyImage_Data, copyUDM2_Mask_Data , unzip\n```\n\n##   Import Module to mosaic and crop raster to Area of Interest using shapefile\n\n```python\nfrom akhdefo_functions import Mosaic, Crop_to_AOI\n```\n\n##  Import Module to filter rasters\n\n```python\nfrom akhdefo_functions import Filter_PreProcess\n```\n\n## Import Module to calculate triplet raster velocities\n\n```python\n from akhdefo_functions import DynamicChangeDetection\n \n ```\n\n##  Import Modules for plotting\n\n```python\nfrom akhdefo_functions import akhdefo_viewer\nfrom akhdefo_functions import plot_stackNetwork\nfrom akhdefo_functions import Akhdefo_resample\nfrom akhdefo_functions import Akhdefo_inversion\n```\n\n##  Import Module to process stacked triplet velocities and collect velocity candidate points\n\n```python\nfrom akhdefo_functions import stackprep'\n```\n##  Python module to coregister satallite images\n\n```python\nfrom akhdefo_functions  import  Coregistration\n```\n\n##  python module to process and calculate timestamp linear deformation rates\n\n```python\nfrom akhdefo_functions import  Time_Series\n```\n\n##  python module to plot timeseries profile for selected points\n\n```python\nfrom  akhdefo_functions import  akhdefo_ts_plot\n\n```",
    'author': 'Mahmud Mustafa Muhammad',
    'author_email': 'mahmud.muhamm1@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/mahmudsfu/AkhDefo',
    'packages': packages,
    'package_data': package_data,
}


setup(**setup_kwargs)
